"""
Example Lambda handler function.
"""
import json
import os
import time

SLEEP = os.getenv('SLEEP') or '0'


def lambda_handler(event, context=None):
    # Do some work...
    time.sleep(float(SLEEP))
    # Return response
    return {
        'body': json.dumps({'text': 'Hello from Lambda Gateway!'}),
        'statusCode': 200,
        'headers': {
            'Content-Type': 'application/json',
        },
    }



import json

from lambda_function import lambda_handler


def test_lambda_handler():
    ret = lambda_handler({})
    exp = {
        'body': json.dumps({'text': 'Hello from Lambda Gateway!'}),
        'statusCode': 200,
        'headers': {
            'Content-Type': 'application/json',
        },
    }
    assert ret == exp



import asyncio
import http
import io
import json
import sys
from unittest import mock

import pytest

from lambda_gateway import server
from lambda_gateway.server import LambdaRequestHandler as Handler


class TestLambdaRequestHandler:
    def setup(self):
        self.subject = mock.Mock(Handler)
        self.subject.base_path = None
        self.subject.timeout = 30
        self.subject.base_path = '/'

        body = json.dumps({'data': 'POST_DATA'})
        self.reqs = {
            'GET': {
                'body': '',
                'httpMethod': 'GET',
                'path': '/',
                'headers': {},
            },
            'HEAD': {
                'body': '',
                'httpMethod': 'HEAD',
                'path': '/',
                'headers': {},
            },
            'POST': {
                'body': body,
                'httpMethod': 'POST',
                'path': '/',
                'headers': {
                    'Content-Length': len(body),
                }
            }
        }

        self.res = {
            'body': json.dumps({'fizz': 'buzz'}),
            'statusCode': 200,
            'headers': {
                'Content-Type': 'application/json',
            },
        }

    def set_request(self, req):
        self.subject.headers = req['headers']
        self.subject.path = req['path']
        self.subject.rfile = io.BytesIO(req['body'].encode())
        self.subject.wfile = io.BytesIO()

    @pytest.mark.parametrize(
        ('base_path', 'exp'),
        [
            (None, None),
            ('simple', 'simple'),
        ],
    )
    def test_set_base_path(self, base_path, exp):
        Handler.set_base_path(base_path)
        assert Handler.base_path == base_path

    def test_set_handler(self):
        handler = 'lambda_function.lambda_handler'
        Handler.set_handler(handler)
        assert Handler.handler == handler

    def test_set_timeout(self):
        Handler.set_timeout(11)
        assert Handler.timeout == 11

    @pytest.mark.parametrize('verb', ['GET', 'HEAD', 'POST'])
    def test_do(self, verb):
        func = getattr(Handler, f'do_{verb}')
        func(self.subject)
        self.subject.invoke.assert_called_once_with(verb)

    @pytest.mark.parametrize('verb', ['GET', 'HEAD', 'POST'])
    def test_get_body(self, verb):
        req = self.reqs[verb]
        self.set_request(req)
        ret = Handler.get_body(self.subject)
        assert ret == req['body']

    @pytest.mark.parametrize('verb', ['GET', 'HEAD', 'POST'])
    def test_get_event(self, verb):
        req = self.reqs[verb]
        self.set_request(req)
        self.subject.get_body.return_value = req['body']
        ret = Handler.get_event(self.subject, verb)
        assert ret == req

    @pytest.mark.parametrize('verb', ['GET', 'HEAD', 'POST'])
    def test_invoke_async(self, verb):
        req = self.reqs[verb]
        self.set_request(req)

        with mock.patch('lambda_gateway.server.get_handler') as mock_handler:

            def handler(event, context=None):
                return self.res

            mock_handler.return_value = handler
            ret = asyncio.run(Handler.invoke_async(self.subject, req))
            assert ret == self.res

    @pytest.mark.parametrize('verb', ['GET', 'HEAD', 'POST'])
    def test_invoke_async_err(self, verb):
        req = self.reqs[verb]
        self.set_request(req)

        with mock.patch('lambda_gateway.server.get_handler') as mock_handler:
            mock_handler.side_effect = Exception
            ret = asyncio.run(Handler.invoke_async(self.subject, req))
            assert ret == server.get_json_response(
                req, 502, message='Internal server error')

    @pytest.mark.parametrize('verb', ['GET', 'HEAD', 'POST'])
    def test_invoke_async_base_path(self, verb):
        self.subject.base_path = '/simple'
        req = self.reqs[verb]
        self.set_request(req)

        with mock.patch('lambda_gateway.server.get_handler'):
            ret = asyncio.run(Handler.invoke_async(self.subject, req))
            exp = server.get_json_response(req, 403, message='Forbidden')
            assert ret == exp

    @pytest.mark.parametrize('verb', ['GET', 'HEAD', 'POST'])
    def test_invoke_with_timeout(self, verb):
        req = self.reqs[verb]
        self.set_request(req)
        asyncio.run(Handler.invoke_with_timeout(self.subject, req))
        self.subject.invoke_async.assert_awaited_once_with(req, None)

    @pytest.mark.parametrize('verb', ['GET', 'HEAD', 'POST'])
    def test_invoke_with_timeout_err(self, verb):
        req = self.reqs[verb]
        self.set_request(req)
        exp = server.get_json_response(
            req, 504, message='Endpoint request timed out')
        self.subject.invoke_async.side_effect = asyncio.TimeoutError
        exe = Handler.invoke_with_timeout(self.subject, req, None)
        ret = asyncio.run(exe)
        assert ret == exp

    # @pytest.mark.skip()
    @pytest.mark.parametrize('verb', ['GET', 'HEAD', 'POST'])
    def test_invoke(self, verb):
        req = self.reqs[verb]
        self.set_request(req)
        self.subject.invoke_with_timeout.return_value = self.res
        Handler.invoke(self.subject, verb)
        self.subject.get_event.assert_called_once_with(verb)
        self.subject.send_response.assert_called_once_with(
            self.res['statusCode'],
        )
        self.subject.send_header.assert_has_calls([
            mock.call(*x) for x in self.res['headers'].items()
        ])
        self.subject.end_headers.assert_called_once_with()
        self.subject.wfile.seek(0)
        ret = self.subject.wfile.read().decode()
        exp = self.res['body']
        assert ret == exp


def test_get_opts_default():
    sys.argv = [
        'lambda-gateway',
        'index.handler',
    ]
    opts = server.get_opts()
    assert opts.bind is None
    assert opts.port == 8000
    assert opts.timeout is None
    assert opts.HANDLER == 'index.handler'


def test_get_opts():
    sys.argv = [
        'lambda-gateway',
        '-b', '0.0.0.0',
        '-p', '8888',
        '-t', '3',
        'index.handler',
    ]
    opts = server.get_opts()
    assert opts.bind == '0.0.0.0'
    assert opts.port == 8888
    assert opts.timeout == 3
    assert opts.HANDLER == 'index.handler'


@pytest.mark.parametrize(
    ('signature', 'exp'),
    [
        ('lambda_function.lambda_handler', 'lambda_handler'),
    ]
)
def test_get_handler(signature, exp):
    ret = server.get_handler(signature).__name__
    assert ret == exp


def test_get_handler_bad_sig():
    with pytest.raises(ValueError):
        server.get_handler('lambda_function')


def test_get_handler_no_file():
    with pytest.raises(ValueError):
        server.get_handler('not_a_file.handler')


def test_get_handler_no_handler():
    with pytest.raises(ValueError):
        server.get_handler('lambda_function.not_a_function')


@pytest.mark.parametrize('base_path', [None, 'simple'])
def test_setup(base_path):
    server.setup('0.0.0.0', 8000, base_path, 30, 'index.handler')
    assert Handler.timeout == 30
    assert Handler.handler == 'index.handler'


@mock.patch('http.server.ThreadingHTTPServer')
def test_run(mock_httpd):
    mock_httpd.socket.getsockname.return_value = ['host', 8000]
    server.run(mock_httpd)
    mock_httpd.serve_forever.assert_called_once_with()


@mock.patch('http.server.ThreadingHTTPServer')
def test_run_int(mock_httpd):
    mock_httpd.socket.getsockname.return_value = ['host', 8000]
    mock_httpd.serve_forever.side_effect = KeyboardInterrupt
    server.run(mock_httpd)
    mock_httpd.serve_forever.assert_called_once_with()
    mock_httpd.shutdown.assert_called_once_with()


@mock.patch('http.server.ThreadingHTTPServer.__enter__')
@mock.patch('lambda_gateway.server.run')
@mock.patch('lambda_gateway.server.setup')
@mock.patch('lambda_gateway.server.get_opts')
def test_main(mock_opts, mock_setup, mock_run, mock_httpd):
    mock_opts.return_value.base_path = 'simple'
    mock_httpd.return_value = '<httpd>'
    mock_setup.return_value = (
        http.server.ThreadingHTTPServer,
        ('localhost', 8000),
        Handler,
    )
    server.main()
    mock_run.assert_called_once_with('<httpd>', '/simple/')



import uuid
from contextlib import contextmanager
from datetime import datetime


@contextmanager
def start(timeout=None):
    """ Yield mock Lambda context object. """
    yield Context(timeout)


class Context:
    """ Mock Lambda context object.

        :param int timeout: Lambda timeout in seconds
    """
    def __init__(self, timeout=None):
        self._start = datetime.utcnow()
        self._timeout = timeout or 30

    @property
    def function_name(self):
        return 'lambda-gateway'

    @property
    def function_version(self):
        return '$LATEST'

    @property
    def invoked_function_arn(self):
        account_id = '123456789012'
        region = 'us-east-1'
        function_name = self.function_name
        return f'arn:aws:lambda:{region}:{account_id}:function:{function_name}'

    @property
    def memory_limit_in_mb(self):
        return 128

    @property
    def aws_request_id(self):
        return str(uuid.uuid1())

    @property
    def log_group_name(self):
        return '/aws/lambda/lambda-gateway'

    @property
    def log_stream_name(self):
        return str(uuid.uuid1())

    def get_remaining_time_in_millis(self):
        delta = datetime.utcnow() - self._start
        remaining_time_in_s = self._timeout - delta.total_seconds()
        if remaining_time_in_s < 0:
            return 0
        return remaining_time_in_s * 1000



from datetime import timedelta

from lambda_gateway import lambda_context


def test_start():
    with lambda_context.start(11) as ret:
        assert isinstance(ret, lambda_context.Context)
        assert ret._timeout == 11


def test_function_name():
    assert lambda_context.Context().function_name == 'lambda-gateway'


def test_function_version():
    assert lambda_context.Context().function_version == '$LATEST'


def test_invoked_function_arn():
    assert lambda_context.Context().invoked_function_arn == \
        'arn:aws:lambda:us-east-1:123456789012:function:lambda-gateway'


def test_memory_limit_in_mb():
    assert lambda_context.Context().memory_limit_in_mb == 128


def test_aws_request_id():
    assert lambda_context.Context().aws_request_id is not None


def test_log_group_name():
    assert lambda_context.Context().log_group_name == \
        '/aws/lambda/lambda-gateway'


def test_log_stream_name():
    assert lambda_context.Context().log_stream_name is not None


def test_get_remaining_time_in_millis():
    context = lambda_context.Context(1)
    assert 0 < context.get_remaining_time_in_millis() < 1000
    context._start -= timedelta(seconds=1)
    assert context.get_remaining_time_in_millis() == 0



#!/usr/bin/env python3
# usage:
#   python server.py --help
import argparse
import asyncio
import importlib
import json
import os
import sys
from http import server

from lambda_gateway import lambda_context


class LambdaRequestHandler(server.SimpleHTTPRequestHandler):
    base_path = None
    timeout = 30

    def do_GET(self):
        self.invoke('GET')

    def do_HEAD(self):
        self.invoke('HEAD')

    def do_POST(self):
        self.invoke('POST')

    def get_body(self):
        """ Get request body to forward to Lambda handler. """
        try:
            content_length = int(self.headers.get('Content-Length'))
            return self.rfile.read(content_length).decode()
        except TypeError:
            return ''

    def get_event(self, httpMethod):
        """ Get Lambda input event object.

            :param str httpMethod: HTTP request method
            :return dict: Lambda event object
        """
        return {
            'body': self.get_body(),
            'headers': dict(self.headers),
            'httpMethod': httpMethod,
            'path': self.path,
        }

    async def invoke_async(self, event, context=None):
        """ Wrapper to invoke the Lambda handler asynchronously.

            :param dict event: Lambda event object
            :param Context context: Mock Lambda context
            :returns dict: Lamnda invocation result
        """
        # Reject request if not starting at base path
        if not self.path.startswith(self.base_path):
            return get_json_response(event, 403, message='Forbidden')

        # Get & invoke Lambda handler
        try:
            handler = get_handler(self.handler)
            loop = asyncio.get_running_loop()
            return await loop.run_in_executor(None, handler, event, context)
        except Exception as err:
            sys.stderr.write(f'{err}\n')
            return get_json_response(
                event, 502, message='Internal server error')

    async def invoke_with_timeout(self, event, context=None):
        """ Wrapper to invoke the Lambda handler with a timeout.

            :param dict event: Lambda event object
            :param Context context: Mock Lambda context
            :returns dict: Lamnda invocation result or 408 TIMEOUT
        """
        try:
            coroutine = self.invoke_async(event, context)
            return await asyncio.wait_for(coroutine, self.timeout)
        except asyncio.TimeoutError:
            return get_json_response(
                event, 504, message='Endpoint request timed out')

    def invoke(self, httpMethod):
        """ Proxy requests to Lambda handler

            :param dict event: Lambda event object
            :param Context context: Mock Lambda context
            :returns dict: Lamnda invocation result
        """
        # Get Lambda event
        event = self.get_event(httpMethod)

        # Get Lambda result
        with lambda_context.start(self.timeout) as context:
            res = asyncio.run(self.invoke_with_timeout(event, context))

        # Parse response
        status = res.get('statusCode', 500)
        headers = res.get('headers', {})
        body = res.get('body', '')

        # Send response
        self.send_response(status)
        for key, val in headers.items():
            self.send_header(key, val)
        self.end_headers()
        self.wfile.write(body.encode())

    @classmethod
    def set_base_path(cls, base_path):
        """ Set base path for REST API.

            :param function handler: Lambda handler function
        """
        cls.base_path = base_path

    @classmethod
    def set_handler(cls, handler):
        """ Set Lambda handler for server.

            :param function handler: Lambda handler function
        """
        cls.handler = handler

    @classmethod
    def set_timeout(cls, timeout):
        """ Set Lambda context timeout.

            :param int timouet: Lambda context timeout in seconds
        """
        cls.timeout = timeout


def get_json_response(event, statusCode, **kwargs):
    if event['httpMethod'] in ['HEAD']:
        body = ''
        size = 0
    else:
        body = json.dumps(kwargs)
        size = len(body)
    return {
        'body': body,
        'statusCode': statusCode,
        'headers': {
            'Content-Type': 'application/json',
            'Content-Length': size,
        },
    }


def get_opts():
    """ Get CLI options. """
    parser = argparse.ArgumentParser(
        description='Start a simple Lambda Gateway server',
    )
    parser.add_argument(
        '-B', '--base-path',
        dest='base_path',
        help='Set base path for REST API',
    )
    parser.add_argument(
        '-b', '--bind',
        dest='bind',
        metavar='ADDRESS',
        help='Specify alternate bind address [default: all interfaces]',
    )
    parser.add_argument(
        '-p', '--port',
        dest='port',
        default=8000,
        help='Specify alternate port [default: 8000]',
        type=int,
    )
    parser.add_argument(
        '-t', '--timeout',
        dest='timeout',
        help='Lambda timeout.',
        type=int,
    )
    parser.add_argument(
        'HANDLER',
        help='Lambda handler signature',
    )
    return parser.parse_args()


def get_handler(signature):
    """ Load handler function.

        :param str signature: Lambda handler signature (eg, 'index.handler')
        :returns function: Lambda handler function
    """
    *path, func = signature.split('.')
    name = '.'.join(path)
    if not name:
        raise ValueError(f"Bad handler signature '{signature}'")
    try:
        pypath = os.path.join(os.path.curdir, f'{name}.py')
        spec = importlib.util.spec_from_file_location(name, pypath)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        return getattr(module, func)
    except FileNotFoundError:
        raise ValueError(f"Unable to import module '{name}'")
    except AttributeError:
        raise ValueError(f"Handler '{func}' missing on module '{name}'")


def setup(bind, port, base_path, timeout, handler):
    """ Setup server.

        :param str bind: Bind address
        :param int port: Port number
        :param str base_path: REST API base path
        :param int timeout: Lambda hanlder timeout in seconds
        :param str handler: Lambda handler signature
    """
    address_family, addr = server._get_best_family(bind, port)
    LambdaRequestHandler.set_base_path(base_path)
    LambdaRequestHandler.set_handler(handler)
    LambdaRequestHandler.set_timeout(timeout)
    server.ThreadingHTTPServer.address_family = address_family
    return (server.ThreadingHTTPServer, addr, LambdaRequestHandler)


def run(httpd, base_path='/'):
    """ Run Lambda Gateway server.

        :param object httpd: ThreadingHTTPServer instance
        :param str base_path: REST API base path
    """
    host, port = httpd.socket.getsockname()[:2]
    url_host = f'[{host}]' if ':' in host else host
    sys.stderr.write(
        f'Serving HTTP on {host} port {port} '
        f'(http://{url_host}:{port}{base_path}) ...\n'
    )
    try:
        httpd.serve_forever()
    except KeyboardInterrupt:
        sys.stderr.write('\nKeyboard interrupt received, exiting.\n')
    finally:
        httpd.shutdown()


def main():
    """ Main entrypoint. """
    opts = get_opts()
    base_path = os.path.join('/', str(opts.base_path or ''), '')
    bind = opts.bind
    port = opts.port
    timeout = opts.timeout
    handler = opts.HANDLER
    HTTPServer, addr, Handler = setup(bind, port, base_path, timeout, handler)
    with HTTPServer(addr, Handler) as httpd:
        run(httpd, base_path)


if __name__ == '__main__':  # pragma: no cover
    main()



from setuptools import setup
from setuptools import find_packages

with open('README.md', 'r') as readme:
    long_description = readme.read()

setup(
    author='amancevice',
    author_email='smallweirdnum@gmail.com',
    description='Simple HTTP server to invoke a Lambda function locally',
    entry_points={
        'console_scripts': [
            'lambda-gateway=lambda_gateway.server:main',
        ],
    },
    install_requires=[],
    long_description=long_description,
    long_description_content_type="text/markdown",
    name='lambda-gateway',
    packages=find_packages(exclude=['tests']),
    python_requires='>=3.6',
    setup_requires=['setuptools_scm'],
    tests_require=[
        'flake8',
        'pytest',
        'pytest-cov',
    ],
    url='https://github.com/amancevice/python-lambda-gateway',
    use_scm_version=True,
)
